@@ -17,6 +17,7 @@ $Config{'startperl'} -w
 !GROK!THIS!
  
 print OUT <<'!NO!SUBS!';
+#!/usr/bin/perl
 #
 # checkbot - A perl5 script to check validity of links in www document trees
 #
@@ -27,7 +28,7 @@ print OUT <<'!NO!SUBS!';
 #
 # Info-URL: http://www.xs4all.nl/~graaff/checkbot/
 #
-# $Id: checkbot.pl,v 1.52 1998/10/10 08:41:50 graaff Exp $
+# $Id: checkbot.pl,v 1.55 1999/05/09 15:30:34 graaff Exp $
 # (Log information can be found at the end of the script)
 
 require 5.001;
@@ -55,7 +56,7 @@ checkbot [B<--debug>] [B<--help>] [B<--verbose>] [B<--url> start URL]
 Checkbot verifies the links in a specific portion of the World Wide
 Web. It creates HTML pages with diagnostics. 
 
-Checkbot's options are:
+Options for Checkbot are:
 
 =over 4
 
@@ -102,10 +103,6 @@ are queried through this proxy server, because Checkbot assumes all
 internal URLs can be accessed directly.  Currently only the HTTP and FTP 
 protocols will be send to the proxy server.
 
-Note that I personally have no use for this feature, and in its current 
-state it is mostly useful for people behind a firewall.  I'd appriciate 
-feedback or comments about how this option could be more useful to you.
-
 =item --mailto <email address>
 
 Send mail to the I<email address> when Checkbot is done
@@ -177,11 +174,16 @@ use AnyDBM_File;
 
 # Declare some global variables, avoids ugly use of main:: all around
 my @starturls = ();
+# A hash of warnings generated by Checkbot apart from HTTP responses
+my %warning = ();
 
+# This hash is indexed by two fields, and contains an array of responses:
+# $problems{HTTP Response code}{URL} = @( HTTP::Response );
+my %problems = ();
 
 # Version information
 my $VERSION;
-( $VERSION ) = sprintf("%d.%02d", q$Revision: 1.52 $ =~ /(\d+)\.(\d+)/);
+( $VERSION ) = sprintf("%d.%02d", q$Revision: 1.55 $ =~ /(\d+)\.(\d+)/);
 
 &check_options();
 &init_modules();
@@ -243,6 +245,8 @@ sub init_modules {
   URI::URL::strict(0);
 
   require Mail::Send if defined $main::opt_mailto;
+
+  use HTTP::Status;
 }
 
 sub init_globals {
@@ -268,7 +272,6 @@ sub init_globals {
 
   # Set up hashes to be used
   %main::checked = ();
-  %main::problem = ();
   %main::servers = ();
 
   # Initialize the start URLs. --url takes precedence. Otherwise
@@ -309,6 +312,11 @@ sub setup {
   mkdir $main::tmpdir, 0755
     || die "$0: unable to create directory $main::tmpdir: $!\n";
 
+  # Explicitly set the record separator. I had the problem that this
+  # was not defined under my perl 5.00502. This should fix that, and
+  # not cause problems for older versions of perl.
+  $/ = "\n";
+
   open(EXTERNAL, ">$main::extfile")
     || die "$0: Unable to open EXTERNAL $main::extfile for writing: $!\n";
   open(CURRENT, ">$main::cur_queue")
@@ -402,6 +410,15 @@ sub handle_url {
     } else {
       $type = guess_media_type($url->path);
     }
+
+    # If we are unsure about the type it could well be a special extension
+    # for e.g. on-the-fly HTML, or a script or some such. Should ask the
+    # server.
+    if ($type eq 'application/octet-stream') {
+      $response = performRequest('HEAD', $url, $urlparent, $type);
+      $type = $response->content_type;
+    }
+
     if ($type =~ /html/o 
         && $url->scheme =~ /^(http|file|ftp|gopher)$/o
         && (!defined $main::opt_exclude || $url !~ /$main::opt_exclude/o)) {
@@ -410,13 +427,8 @@ sub handle_url {
       $reqtype = 'HEAD';
     }
 
-    # Output what we are going to do with this link
-    printf STDERR "    %4s %s (%s)\n", $reqtype, $url, $type
-      if $main::opt_verbose;
-	
-    my $ref_header = new HTTP::Headers 'Referer' => $urlparent;
-    my $request = new HTTP::Request($reqtype, $url, $ref_header);
-    my $response = $main::ua->simple_request($request);
+    $response = performRequest($reqtype, $url, $urlparent, $type)
+      unless defined($response) and $reqtype eq 'HEAD';
 
     if ($response->is_success) {
       sleep($main::opt_sleep) unless $main::opt_debug || $url->scheme eq 'file';
@@ -434,13 +446,15 @@ sub handle_url {
       } else {
 	print STDERR "         ", $response->code, ' ', $response->message, "\n" 
 	  if $main::opt_verbose;
-	$main::problem{$url} = $response;
+        push @{$problems{$response->code}{$url}}, $response;
 	$main::st_int[$main::PROBL]++;
       }
 
       if ($response->is_redirect) {
-	print STDERR "         Redirected to " . $response->header('Location') . "\n" if $main::opt_verbose;
-	add_to_queue($response->header('Location'), $urlparent);
+	my $redir_url = new URI::URL $response->header('Location'), $url;
+	print STDERR "         Redirected to " . $redir_url . "\n" if $main::opt_verbose;
+
+	add_to_queue($redir_url, $urlparent);
       }
     }
 
@@ -451,6 +465,20 @@ sub handle_url {
   }
 }
 
+sub performRequest {
+  my ($reqtype, $url, $urlparent, $type) = @_;
+
+  # Output what we are going to do with this link
+  printf STDERR "    %4s %s (%s)\n", $reqtype, $url, $type
+    if $main::opt_verbose;
+	
+  my $ref_header = new HTTP::Headers 'Referer' => $urlparent;
+  my $request = new HTTP::Request($reqtype, $url, $ref_header);
+  my $response = $main::ua->simple_request($request);
+
+  return $response;
+}
+
 sub check_external {
   my $newurl;
   my $urlparent;
@@ -518,7 +546,7 @@ sub check_external {
 	  printf STDERR "          %d (%s)\n", $response->code, 
 	  $response->message
 	    if $main::opt_verbose;
-	  $main::problem{$url} = $response;
+          push @{$problems{$response->code}{$url}}, $response;
 	  $main::st_ext[$main::PROBL]++;
 	}
       }
@@ -594,17 +622,10 @@ sub create_page {
     print OUT "</table>\n\n";
     
     # Server information
-    my $server;
-    print OUT "<hr><h2>Overview per server</h2>\n";
-    print OUT "<table> <tr> <th>Server</th><th>Server<br>Type</th><th>Unique<br>links</th><th>Problem<br>links</th><th>Ratio</th></tr>\n";
-    foreach $server (sort keys %main::servers) {
-      &print_server($server, $final_page);
-    }
-    print OUT "</table>\n\n";
-    print OUT "<hr>\n";
+    printAllServers($final_page);
 
     # Checkbot session parameters
-    print OUT "<h2>Checkbot session parameters</h2>\n";
+    print OUT "<hr><h2>Checkbot session parameters</h2>\n";
     print OUT "<table>\n";
     print OUT "<tr><th align=left>--url</th><td>Start URL(s)</td><td>",
               join(',', @starturls), "</td></tr>\n";
@@ -626,6 +647,28 @@ sub create_page {
     print STDERR "*** Done writing result page\n" if $main::opt_verbose;
 }
 
+# Create a list of all the servers, and create the corresponding table
+# and subpages. We use the servers overview for this. This can result
+# in strange effects when the same server (e.g. IP address) has
+# several names, because several entries will appear. However, when
+# using the IP address there are also a number of tricky situations,
+# e.g. with virtual hosting. Given that likely the servers have
+# different names for a reasons, I think it is better to have
+# duplicate entries in some cases, instead of working off of the IP
+# addresses.
+
+sub printAllServers {
+  my ($finalPage) = @_;
+
+  my $server;
+  print OUT "<hr><h2>Overview per server</h2>\n";
+  print OUT "<table> <tr> <th>Server</th><th>Server<br>Type</th><th>Unique<br>links</th><th>Problem<br>links</th><th>Ratio</th></tr>\n";
+
+  foreach $server (sort keys %main::servers) {
+    print_server($server, $finalPage);
+  }
+  print OUT "</table>\n\n";
+}
 
 sub get_server_type {
   my($server) = @_;
@@ -692,6 +735,13 @@ sub handle_doc {
       # Remove fragments, if any
       $url->frag(undef);
 
+      # Check whether URL has fully-qualified hostname
+      if ($url->scheme =~ /^(http|ftp)/ and $url->host !~ /\./) {
+        print STDERR "--> Unqualified host name: $url\n";
+        $warning{'Unqualified host name in URL'}{$url} .= $url->base . "\n";
+      }
+
+
       if ($url->abs =~ /$main::opt_match/o) {
 	if (defined $main::checked{$url->abs}) {
 	  $doc_dup++;
@@ -730,7 +780,10 @@ sub add_to_queue {
 sub print_server {
   my($server, $final_page) = @_;
 
-  print STDERR "    Writing server $server (really ", ip_address($server), ")\n" if $main::opt_verbose;
+  my $host = $server;
+  $host =~ s/(.*):\d+/$1/;
+
+  print STDERR "    Writing server $server (really ", ip_address($host), ")\n" if $main::opt_verbose;
 
   my $server_problem = &count_problems($server);
   my $filename = "$main::server_prefix-$server.html";
@@ -764,11 +817,14 @@ sub print_server {
   print SERVER "<title>Checkbot: output for server $server</title></head>\n";
   print SERVER "<body><h2><em>Checkbot</em>: report for server <tt>$server</tt></h2>\n";
   print SERVER "Go To: <a href=\"$main::file\">Main report page</a>.\n";
-  print SERVER "<hr><dl>\n";
+  print SERVER "<hr>\n";
 
-  &print_server_problems($server, $final_page);
+  printServerProblems($server, $final_page);
 
-  print SERVER "</dl>\n";
+  print SERVER "<hr>\n";
+  printServerWarnings($server, $final_page);
+
+  print SERVER "\n";
   print SERVER signature();
 
   close SERVER;
@@ -776,41 +832,80 @@ sub print_server {
 
 # Return a string containing Checkbot's signature for HTML pages
 sub signature {
-  return "<hr>\nPage created by <a href=http://www.xs4all.nl/~graaff/checkbot/>Checkbot $VERSION</a> on <em>" . localtime() . "</em>.\n";
+  return "<hr>\nPage created by <a href=\"http://www.xs4all.nl/~graaff/checkbot/\">Checkbot $VERSION</a> on <em>" . localtime() . "</em>.\n</body></html>";
 }
 
-sub print_server_problems {
+# Loop through all possible problems, select relevant ones for this server
+# and display them in a meaningful way.
+sub printServerProblems {
   my ($server) = @_;
 
-  my $prevpath = "";
-  my $prevcode = 0;
-  my $prevmessage = "";
+  foreach my $code (sort keys %problems) {
+    my $codeOutput = '';
+    my %thisServerList = ();
 
-  my $url;
+    foreach my $url (sort keys %{ $problems{$code}}) {
 
-  foreach $url (sort byproblem keys %main::problem) {
-    my $response = $main::problem{$url};
-    my $parent = new URI::URL $response->request->header('Referer');
-    if ($parent =~ $server) {
-      if ($response->is_error || $response->is_redirect) {
-	my $path = '/' . $parent->path;
-	my $string = "";
-	if ($response->code != $prevcode 
-	    || lc($response->message) ne $prevmessage) {
-	  $string .= '</dl><h4><b>' . $response->code . '</b> '
-	    . $response->message . "</h4>\n<dl>\n";
-	  $prevcode = $response->code;
-	  $prevmessage = lc($response->message);
-	  $prevpath = "";
-	}
-	if ($path ne $prevpath) {
-	  $string .= "<dt><b><a href=\"$parent\">$path</a></b>";
-	  $prevpath = $path;
-	}
-	$string .= "<dd><a href=\"$url\">$url</a>\n";
-	print SERVER $string;
+      foreach my $response (@{ $problems{$code}{$url} }) {
+	my $parent = $response->request->header('Referer');
+	next if $parent !~ $server;
+
+	chomp $parent;
+	$thisServerList{$parent}{$url} = $response;
       }
     }
+      
+    foreach my $parent (sort keys %thisServerList) {
+      my $urlOutput = '';
+      foreach my $url (sort keys %{ $thisServerList{$parent} }) {
+	my $response = $thisServerList{$parent}{$url};
+	$urlOutput .= "<li><a href=\"$url\">$url</a><br>\n";
+	$urlOutput .= $response->message . "\n" if $response->message ne status_message($code);
+      }
+      if ($urlOutput ne '') {
+	$codeOutput .= "<dt><a href=\"$parent\">$parent</a>\n<dd>\n";
+	$codeOutput .= "<ul>\n$urlOutput\n</ul>\n\n";
+      }
+    }
+
+    if ($codeOutput ne '') {
+      print SERVER "<h4>$code ", status_message($code), "</h4>\n";
+      print SERVER "<dl>\n$codeOutput\n</dl>\n";
+    }
+  }
+
+}
+
+sub printServerWarnings {
+  my ($server,$finalPage) = @_;
+
+  my $warningType;
+  my $oldWarningType = '';
+  my $url;
+  my $parent;
+
+  foreach $warningType (sort keys %warning) {
+    print SERVER "<h4>$warningType</h4>\n<dl>\n" unless $oldWarningType eq $warningType;
+    $oldWarningType = $warningType;
+    my %thisServerList = ();
+
+    foreach $url (sort keys %{$warning{$warningType}}) {
+      # Get only those parent URLs that match this server
+      my @parents = grep(/$server/, split(/\n/, $warning{$warningType}{$url}));
+      if ($#parents >= 0) {
+	foreach my $parent (@parents) {
+	  $thisServerList{$parent}{$url} = 1;
+	}
+      }
+    }
+
+    foreach my $parent (sort keys %thisServerList) {
+      print SERVER "<dt><a href=\"$parent\">$parent</a>\n<dd>\n";
+      foreach my $url (sort keys %{ $thisServerList{$parent} }) {
+	print SERVER "<a href=\"$url\">$url</a><br>";
+      }
+    }
+    print SERVER "</dl>\n";
   }
 }
 
@@ -898,35 +993,38 @@ sub ip_address {
 sub count_problems {
   my ($server) = @_;
   my $count = 0;
-  my $url = '';
 
-  foreach $url (keys %main::problem) {
-    my $response = $main::problem{$url};
-    my $parent = new URI::URL $response->request->header('Referer');
-    $count++ if $parent =~ $server;
+  foreach my $code (sort keys %problems) {
+    foreach my $url (sort keys %{ $problems{$code}}) {
+      foreach my $response (@{ $problems{$code}{$url} }) {
+	my $parent = new URI::URL $response->request->header('Referer');
+	$count++ if $parent =~ $server;
+      }
+    }
   }
 
   return $count;
 }
 
-sub byproblem {
-    my $response_a = $main::problem{$::a};
-    my $response_b = $main::problem{$::b};
-    
-    if ($response_a->code == $response_b->code) {
-	if ($response_a->message eq $response_b->message) {
-	    return $response_a->request->header('Referer') 
-		cmp $response_b->request->header('Referer');
-	} else {
-	    return $response_a->message cmp $response_b->message;
-	}
-    } else {
-	return $response_a->code <=> $response_b->code;
-    }
-}
-
 
 # $Log: checkbot.pl,v $
+# Revision 1.55  1999/05/09 15:30:34  graaff
+# List broken links under the pages that contain them, instead of the
+# other way around, reverting back to the way things are in 1.53 and
+# earlier.
+# Handle redirected, but unqualified links.
+# Only print each warning header once.
+# Documentation fixes.
+#
+# Revision 1.54  1999/01/18 22:22:29  graaff
+# Fixed counting of problem links to correct checkbot.html results page.
+#
+# Revision 1.53  1999/01/17 20:59:14  graaff
+# Fixed internal problem storage.
+# Changed report to collate HTTP response codes.
+# Added warning section to pages with additional warnings.
+# Hammered out bug with record separator in perl 5.005.
+#
 # Revision 1.52  1998/10/10 08:41:50  graaff
 # new version, some documentation work, and the HTML::Parse problem fixed.
 #
